@@ -3,9 +3,13 @@ package Image::WordCloud::BoundingBox;
 use strict;
 use warnings;
 
-=head1 SUBROUTINES/METHODS
+=head1 SYNOPSIS
 
-=head2 new
+Not used currently
+
+=head2 new()
+
+Create a bounding box
 
 =cut
 
